@@ -8,7 +8,12 @@ from itertools import product
 
 import numpy as np
 
-from .exceptions import AggregationError, DimensionStructureError, DuplicateNameError
+from .exceptions import (
+    AggregationError,
+    ArithmeticError,
+    DimensionStructureError,
+    DuplicateNameError,
+)
 from .functions import create_group_matrix, get_path_up_down, group_unique_values
 
 # pandas is no a required dependency
@@ -259,6 +264,11 @@ class Domain:
         dims_str = ", ".join("%s=%s(%d)" % (n, d, d.size) for n, d in dims)
         return "(" + dims_str + ")"
 
+    def __eq__(self, other):
+        return self._size == other._size and all(
+            d1 == d2 for d1, d2 in zip(self.dimension_levels, other.dimension_levels)
+        )
+
     def dict_to_matrix(self, data):
         """create data matrix from dict data
 
@@ -347,15 +357,158 @@ class Domain:
     def shape(self):
         return self._shape
 
+    @property
+    def is_scalar(self):
+        return self.size == 0
 
-class Variable:
+
+class VariableBase:
+    pass
+
+
+class MixinNumpyMath:
+    """delegate math operations to numpy"""
+
+    def _get_unit_add(self, other):
+        unit = self._unit
+        if isinstance(other, VariableBase):
+            other_unit = other._unit
+        else:
+            other_unit = unit
+        if unit != other_unit:
+            raise ArithmeticError("unit mismatch")
+        return unit
+
+    def _get_unit_mult(self, other):
+        unit = self._unit
+        if isinstance(other, VariableBase):
+            other_unit = other._unit
+        else:
+            other_unit = unit
+        if unit or other_unit:
+            raise NotImplementedError("unit multiplication")
+        return unit
+
+    def _get_unit_div(self, other):
+        unit = self._unit
+        if isinstance(other, VariableBase):
+            other_unit = other._unit
+        else:
+            other_unit = unit
+        if unit or other_unit:
+            raise NotImplementedError("unit division")
+        return unit
+
+    def _get_unit_rdiv(self, other):
+        unit = self._unit
+        if isinstance(other, VariableBase):
+            other_unit = other._unit
+        else:
+            other_unit = unit
+        if unit or other_unit:
+            raise NotImplementedError("unit division")
+        return unit
+
+    def _get_vartype(self, other):
+        vartype = self._vartype
+        if isinstance(other, VariableBase):
+            other_vartype = other._vartype
+        else:
+            other_vartype = vartype
+        if vartype != other_vartype:
+            raise ArithmeticError("vartype mismatch")
+        if vartype != "extensive":
+            raise ArithmeticError("invalid vartype for operation: %s" % vartype)
+        return vartype
+
+    def _get_domain(self, other):
+        # domain must be either equal, or one must be a scalar
+        domain = self._domain
+        if isinstance(other, VariableBase):
+            other_domain = other._domain
+        else:
+            other_domain = domain
+        if domain.is_scalar:
+            return other_domain
+        elif other_domain.is_scalar:
+            return domain
+        if domain != other_domain:
+            raise ArithmeticError("vartype mismatch")
+        return domain
+
+    @staticmethod
+    def _get_data_matrix_other(other):
+        if isinstance(other, VariableBase):
+            data_matrix = other._data_matrix
+        else:
+            data_matrix = other
+        return data_matrix
+
+    def __add__(self, other):
+        unit = self._get_unit_add(other)
+        vartype = self._get_vartype(other)
+        domain = self._get_domain(other)
+        # use numpy to do actual calculation
+        data = self._data_matrix + self._get_data_matrix_other(other)
+        return Variable(data, domain, vartype, unit=unit)
+
+    def __sub__(self, other):
+        unit = self._get_unit_add(other)
+        vartype = self._get_vartype(other)
+        domain = self._get_domain(other)
+        # use numpy to do actual calculation
+        data = self._data_matrix - self._get_data_matrix_other(other)
+        return Variable(data, domain, vartype, unit=unit)
+
+    def __mul__(self, other):
+        unit = self._get_unit_mult(other)
+        vartype = self._get_vartype(other)
+        domain = self._get_domain(other)
+        # use numpy to do actual calculation
+        data = self._data_matrix * self._get_data_matrix_other(other)
+        return Variable(data, domain, vartype, unit=unit)
+
+    def __rmul__(self, other):
+        unit = self._get_unit_mult(other)
+        vartype = self._get_vartype(other)
+        domain = self._get_domain(other)
+        # use numpy to do actual calculation
+        data = self._get_data_matrix_other(other) * self._data_matrix
+        return Variable(data, domain, vartype, unit=unit)
+
+    def __truediv__(self, other):
+        unit = self._get_unit_rdiv(other)
+        vartype = self._get_vartype(other)
+        domain = self._get_domain(other)
+        # use numpy to do actual calculation
+        data = self._data_matrix / self._get_data_matrix_other(other)
+        return Variable(data, domain, vartype, unit=unit)
+
+    def __rtruediv__(self, other):
+        unit = self._get_unit_rdiv(other)
+        vartype = self._get_vartype(other)
+        domain = self._get_domain(other)
+        # use numpy to do actual calculation
+        data = self._get_data_matrix_other(other) / self._data_matrix
+        return Variable(data, domain, vartype, unit=unit)
+
+    # unary methods
+    def __neg__(self):
+        data = -self._data_matrix
+        return Variable(data, self._domain, self._vartype, unit=self._unit)
+
+    def fillna(self, value=0):
+        data = np.nan_to_num(self._data_matrix, nan=value)
+        return Variable(data, self._domain, self._vartype, unit=self._unit)
+
+
+class Variable(VariableBase, MixinNumpyMath):
     """Data container over a multidimensional domain.
 
     Comparable to pandas.Series with multiindex
     and/or multidimensional numpy.ndarray
 
     Args:
-        name(str): name of variable
         data: data can be passed in different ways:
 
           * dictionary of key -> value,
@@ -379,14 +532,13 @@ class Variable:
 
     """
 
-    def __init__(self, name, data, domain, vartype, unit=None):
+    def __init__(self, data, domain, vartype, unit=None):
         self._unit = unit
 
         if vartype not in ("intensive", "extensive", "weight"):
             raise ValueError("vartype not in ('intensive', 'extensive', 'weight')")
 
         self._vartype = vartype
-        self._name = name or "UnnamedVariable"
 
         # set domain
         if not domain:  # Scalar
@@ -430,9 +582,6 @@ class Variable:
             if not np.allclose(np.ones(shape), sums):
                 raise ValueError("Values in some groups don't add up to 1.0")
 
-    def __str__(self):
-        return str(self.name)
-
     def to_series(self):
         """Return indexed pandas Series"""
         if not pd:
@@ -446,12 +595,15 @@ class Variable:
 
         if not self._domain.dimensions:
             index = None
-        else:
+        elif self._domain.size() == 1:  # one dimensional index
+            dimension_level = self._domain.dimension_levels[0]
+            index = pd.Index(dimension_level.elements, name=dimension_level.name)
+        else:  # multilevel index
             index = pd.MultiIndex.from_tuples(
                 self._domain._keys, names=self._domain.dimension_level_names
             )
 
-        return pd.Series(data, index=index, name=self.name)
+        return pd.Series(data, index=index)
 
     def to_dict(self, skip_0=False):
         """Return data as an dict
@@ -482,12 +634,9 @@ class Variable:
 
         return res
 
-    def to_weight(self, name=None):
+    def to_weight(self):
         """Convert variable into weight.
         Only possible if Domain has only one dimension.
-
-        Args:
-            name(str, optional): new name
         """
 
         if self._domain.size != 1:
@@ -511,21 +660,19 @@ class Variable:
 
         # is_intensive=None ==> weights
         return Variable(
-            name=name or self.name,
             domain=self._domain,
             data=data,
             unit=None,
             vartype="weight",
         )
 
-    def aggregate(self, dimension_name, weights=None, name=None):
+    def aggregate(self, dimension_name, weights=None):
         """aggregate one dimension lvel to the next
 
         Args:
             dimension_name(str): name of dimension to be aggregated
             weights(Weights, optional): weights variable for aggregation,
               must have the the target dimension level as domain
-            name(str, optional): new name
         """
 
         # get the current level of the dimension
@@ -582,16 +729,13 @@ class Variable:
         ).swapaxes(dim_idx_last, dim_idx)
 
         return Variable(
-            name=name or self.name,
             domain=new_domain,
             data=data_matrix,
             unit=self._unit,
             vartype=self._vartype,
         )
 
-    def disaggregate(
-        self, dimension_name, dimension_level_name, weights=None, name=None
-    ):
+    def disaggregate(self, dimension_name, dimension_level_name, weights=None):
         """disaggregate one dimension to one of the next child levels
 
         Args:
@@ -599,7 +743,6 @@ class Variable:
             dimension_level_name(str): name of new child level
             weights(Weights, optional): weights variable for disaggregation,
               must have the the target dimension level as domain
-            name(str, optional): new name
         """
 
         if self.is_extensive and not weights:
@@ -668,19 +811,17 @@ class Variable:
         ).swapaxes(dim_idx_last, dim_idx)
 
         return Variable(
-            name=name or self.name,
             domain=new_domain,
             data=data_matrix,
             unit=self._unit,
             vartype=self._vartype,
         )
 
-    def expand(self, dimension, name=None):
+    def expand(self, dimension):
         """add dimension on aggregated level
 
         Args:
             dimension(Dimension): root level dimension
-            name(str, optional): new name
         """
 
         if not dimension.is_dimension_root:
@@ -694,19 +835,17 @@ class Variable:
         data_matrix = np.expand_dims(self._data_matrix, axis=self._domain.size)
         domain = Domain(list(self._domain.dimension_levels) + [dimension])
         return Variable(
-            name=name or self.name,
             domain=domain,
             data=data_matrix,
             unit=self._unit,
             vartype=self._vartype,
         )
 
-    def squeeze(self, dimension_name, name=None):
+    def squeeze(self, dimension_name):
         """remove aggregated dimension
 
         Args:
             dimension_name: names of existing dimensions to be dropped
-            name(str, optional): new name
         """
 
         logging.debug("removing %s" % (dimension_name,))
@@ -724,20 +863,18 @@ class Variable:
             + self._domain.dimension_levels[dim_idx + 1 :]
         )
         return Variable(
-            name=name or self.name,
             domain=domain,
             data=data_matrix,
             unit=self._unit,
             vartype=self._vartype,
         )
 
-    def reorder(self, dimension_names, name=None):
+    def reorder(self, dimension_names):
         """Return new Variable with reordered Dimensions
 
         Args:
             dimension_names(list): list of names of existing dimensions
               in desired order
-            name(str, optional): new name
         """
         if set(dimension_names) != set(self._domain.dimension_names):
             raise DimensionStructureError(
@@ -748,7 +885,6 @@ class Variable:
         data_matrix = self._data_matrix.transpose(indices)
         domain = Domain([self._domain.dimension_levels[i] for i in indices])
         return Variable(
-            name=name or self.name,
             domain=domain,
             data=data_matrix,
             unit=self._unit,
@@ -845,9 +981,13 @@ class Variable:
             if squeeze:
                 steps.append((dim, None, "squeeze", None))
 
+            if not steps:
+                # no change: still add a dummy step (for visualization)
+                steps.append((source_level, source_level, "keep", None))
+
         return result
 
-    def transform(self, domain, level_weights=None, name=None):
+    def transform(self, domain, level_weights=None):
         """Main function to map variable to a new domain.
 
         Args:
@@ -855,8 +995,6 @@ class Variable:
             level_weights(dict, optional):
                dimension level names -> one dimensional variables that will be used
                as weights for this level
-            name(str, optional): new name
-
         """
 
         if not isinstance(domain, Domain):
@@ -879,16 +1017,15 @@ class Variable:
                     result = result.disaggregate(
                         dim_name, to_level.name, weights=weight
                     )
+                elif action == "keep":
+                    # do nothing
+                    pass
                 else:
                     raise NotImplementedError(action)
 
-        result = result.reorder(domain.dimension_names, name=name)
+        result = result.reorder(domain.dimension_names)
         return result
 
-    @property
-    def name(self):
-        return self._name
-
     @property
     def is_intensive(self):
         return self._vartype == "intensive"
@@ -903,14 +1040,13 @@ class Variable:
 
     @property
     def is_scalar(self):
-        return self._domain.size == 0
+        return self._domain.is_scalar
 
 
 class ExtensiveVariable(Variable):
     """Shorthand for `Variable(vartype="extensive")`
 
     Args:
-        name(str): name of variable
         data: data can be passed in different ways:
 
           * dictionary of key -> value, where key is a tuple that is an element
@@ -925,17 +1061,14 @@ class ExtensiveVariable(Variable):
 
     """
 
-    def __init__(self, name, data, domain, unit=None):
-        super().__init__(
-            name=name, data=data, unit=unit, domain=domain, vartype="extensive"
-        )
+    def __init__(self, data, domain, unit=None):
+        super().__init__(data=data, unit=unit, domain=domain, vartype="extensive")
 
 
 class IntensiveVariable(Variable):
     """Shorthand for `Variable(vartype="intensive")`
 
     Args:
-        name(str): name of variable
         data: data can be passed in different ways:
 
           * dictionary of key -> value,
@@ -950,10 +1083,8 @@ class IntensiveVariable(Variable):
 
     """
 
-    def __init__(self, name, data, domain, unit=None):
-        super().__init__(
-            name=name, data=data, unit=unit, domain=domain, vartype="intensive"
-        )
+    def __init__(self, data, domain, unit=None):
+        super().__init__(data=data, unit=unit, domain=domain, vartype="intensive")
 
 
 class Weight(Variable):
@@ -965,7 +1096,6 @@ class Weight(Variable):
     * values in each group add up to 1.0
 
     Args:
-        name(str): name of variable
         data: data can be passed in different ways:
 
           * dictionary of key -> value,
@@ -979,9 +1109,9 @@ class Weight(Variable):
 
     """
 
-    def __init__(self, name, data, dimension_level):
+    def __init__(self, data, dimension_level):
         super().__init__(
-            name=name, data=data, unit=None, domain=[dimension_level], vartype="weight"
+            data=data, unit=None, domain=[dimension_level], vartype="weight"
         )
 
 
@@ -989,28 +1119,26 @@ class ExtensiveScalar(ExtensiveVariable):
     """Shorthand for `ExtensiveVariable(domain=None)`
 
     Args:
-        name(str): name of variable
         value: number
         unit(optional): not implemented yet
 
     """
 
-    def __init__(self, name, value, unit=None):
-        super().__init__(name=name, data=value, unit=unit, domain=[])
+    def __init__(self, value, unit=None):
+        super().__init__(data=value, unit=unit, domain=[])
 
 
 class IntensiveScalar(IntensiveVariable):
     """Shorthand for `IntensiveVariable(domain=None)`
 
     Args:
-        name(str): name of variable
         value: number
         unit(optional): not implemented yet
 
     """
 
-    def __init__(self, name, value, unit=None):
-        super().__init__(name=name, data=value, unit=unit, domain=[])
+    def __init__(self, value, unit=None):
+        super().__init__(data=value, unit=unit, domain=[])
 
 
 if __name__ == "__main__":
